#include "BigInteger.hpp"

#define PI 3.14159265358979323846

BigInteger::BigInteger() : BigInteger("0")
{
}

BigInteger::BigInteger(const BigInteger &number)
{
	sign = number.sign;
	len = number.len;
	cap = number.cap;
	digits = new char[cap];

	for (size_t i = 0; i < len; i++)
	{
		digits[i] = number.digits[i];
	}
}

BigInteger::BigInteger(std::string number) : BigInteger(number, 32)
{
}

BigInteger::BigInteger(std::string number, size_t capacity)
{
	len = 0;
	cap = capacity;
	sign = 1;
	digits = new char[cap];

	size_t index = 0;

	if (number.length() == 0)
	{
		digits[0] = 0;
		len = 1;
		return;
	}

	if (number[0] == '-')
	{
		sign = -1;
		index++;
	}
	else if (number[0] == '+')
	{
		index++;
	}

	while (index < number.length() && number[index] >= '0' && number[index] <= '9')
	{
		len++;
		index++;
	}

	if (len == 0)
	{
		digits[0] = 0;
		len = 1;
		sign = 1;
	}
	else
	{
		while (cap < len)
			cap <<= 1;
		delete[] digits;
		digits = new char[cap];
		for (size_t i = 0; i < len; i++)
		{
			digits[i] = number[--index] - '0';
		}
	}
}

BigInteger::~BigInteger()
{
	if (digits != nullptr)
	{
		delete[] digits;
	}
}

size_t BigInteger::Length() const
{
	return len;
}

BigInteger BigInteger::operator*(const BigInteger &number) const
{
	size_t n = 1;
	while (n < number.len + len)
		n <<= 1;

	std::vector<std::complex<double>> a;
	std::vector<std::complex<double>> b;

	// 用两数填满长度为n的vector中，即得到最高次数n-1的多项式系数向量，高次项不足处补0
	for (size_t i = 0; i < len; i++)
		a.push_back(digits[i]);
	for (size_t i = 0; i < number.len; i++)
		b.push_back(number.digits[i]);

	while (a.size() < n)
		a.push_back(0);
	while (b.size() < n)
		b.push_back(0);

	// 离散傅里叶变换
	a = fft(a);
	b = fft(b);

	// 时域相乘，结果存到a中
	for (size_t i = 0; i < n; i++)
	{
		a[i] *= b[i];
	}

	// 离散傅里叶逆变换
	a = ifft(a);

	// 将变换结果转换为BigInteger
	BigInteger result("0", number.len + len);
	int carry = 0;
	for (size_t i = 0; i < number.len + len; i++)
	{
		int temp = (int)(a[i].real() + 0.5) + carry;
		result.digits[i] = temp % 10;
		carry = temp / 10;
	}

	result.len = number.len + len;
	while (result.len > 1 && result.digits[result.len - 1] == 0)
		result.len--;
	result.sign = number.sign * sign;

	return result;
}

BigInteger &BigInteger::operator=(const BigInteger &number)
{
	sign = number.sign;
	len = number.len;

	if (len > cap)
	{
		delete[] digits;
		cap = number.cap;
		digits = new char[cap];
	}

	for (size_t i = 0; i < len; i++)
	{
		digits[i] = number.digits[i];
	}

	return *this;
}

bool BigInteger::operator==(const BigInteger &number) const
{
	if (len != number.len)
		return false;
	if (len == 1 && digits[0] == 0 && number.digits[0] == 0)
		return true;
	if (sign != number.sign)
		return false;

	for (size_t i = 0; i < len; i++)
	{
		if (digits[i] != number.digits[i])
			return false;
	}

	return true;
}

std::istream &operator>>(std::istream &in, BigInteger &number)
{
	std::string str;
	in >> str;
	number = BigInteger(str);

	return in;
}

std::ostream &operator<<(std::ostream &out, const BigInteger &number)
{
	if (number.len == 1 && number.digits[0] == 0)
	{
		out << "0";
		return out;
	}

	std::string str;
	if (number.sign < 0)
	{
		str = "-";
	}

	for (size_t i = number.len; i != 0; i--)
	{
		str += number.digits[i - 1] + '0';
	}

	out << str;
	return out;
}

void BigInteger::resize(size_t size)
{
	if (cap >= size)
		return;

	char *temp = new char[size];
	for (size_t i = 0; i < len; i++)
	{
		temp[i] = digits[i];
	}

	delete[] digits;
	digits = temp;
}

std::vector<std::complex<double>> BigInteger::fft(const std::vector<std::complex<double>> &factors) const
{
	return fft_algorithm(factors);
}

std::vector<std::complex<double>> BigInteger::ifft(const std::vector<std::complex<double>> &factors) const
{
	std::vector<std::complex<double>> result = fft_algorithm(factors, true);
	for (size_t i = 0; i < result.size(); i++)
	{
		result[i] /= result.size();
	}

	return result;
}

std::vector<std::complex<double>> BigInteger::fft_algorithm(const std::vector<std::complex<double>> &factors, bool inverse) const
{
	if (factors.size() == 1)
		return factors;

	std::vector<std::complex<double>> odd; // 奇数项系数/结果
	std::vector<std::complex<double>> even; // 偶数项系数/结果

	for (size_t i = 0; i < factors.size(); i++)
	{
		if (i % 2 == 1)
			odd.push_back(factors[i]);
		else
			even.push_back(factors[i]);
	}

	// 递归分别求奇偶项的离散傅里叶变换
	odd = fft_algorithm(odd, inverse);
	even = fft_algorithm(even, inverse);
	size_t n = factors.size();
	std::vector<std::complex<double>> result;

	for (size_t i = 0; i < n; i++)
	{
		std::complex<double> w(cos(2 * PI * i / n), sin(2 * PI * i / n));
		if (inverse) w = std::conj(w);
		result.push_back(even[i % (n / 2)] + w * odd[i % (n / 2)]);
	}

	return result;
}